#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Any, Dict, List, Union
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.schema import LLMResult

import threading
import asyncio
import re
from ..custom_logging import log

class ContentBuffer:
    """
    A buffer class for storing and managing textual content.

    This class provides methods to write text to the buffer, read the entire buffer content,
    and clear the buffer content. The buffer can be used to collect text output for further
    processing or inspection.

    Attributes:
        content (str): Stores the textual content of the buffer.
    """

    def __init__(self):
        """
        Initializes a new ContentBuffer instance.

        The content buffer starts with an empty string, and logging is initialized to indicate
        that the buffer has been created.
        """
        self.content = ""
        self.lock = asyncio.Lock()  # Use an async lock to ensure atomic writes
        self.content_available = asyncio.Event()
        log.debug("Content buffer initialized")
    
    def write(self, text: str):
        """
        Writes text to the content buffer.

        Args:
            text (str): The text to be added to the buffer.

        Adds the given text to the existing content of the buffer.
        """
        self.content += text

    async def async_write(self, text: str):
        """
        Asynchronously writes text to the content buffer.

        Args:
            text (str): The text to be added to the buffer.

        Adds the given text to the existing content of the buffer.
        """
        async with self.lock:
            self.content += text
            self.content_available.set()
    
    def read(self) -> str:
        """
        Reads the entire content from the buffer.

        Returns:
            str: The content of the buffer.

        Provides the entire content stored in the buffer.
        """   
        return self.content

    async def async_read(self) -> str:
        """
        Asynchronously reads the entire content from the buffer.

        Returns:
            str: The content of the buffer.
        """
        async with self.lock:
            content = self.content
            return content

    def clear(self):
        """
        Clears the content buffer.

        Empties the buffer content, resetting it to an empty string.
        """
        self.content = ""

    async def async_clear(self):
        """
        Asynchronously clears the content buffer.

        Empties the buffer content, resetting it to an empty string.
        """
        async with self.lock:
            self.content = ""
            self.content_available.clear()


class BufferStreamingStdOutCallbackHandler(StreamingStdOutCallbackHandler):
    """
    A callback handler for streaming LLM output to a content buffer.

    This class handles the streaming of output from a large language model (LLM),
    processes tokens from the model output, and writes them to a ContentBuffer.
    It supports handling different types of tokens and keeps track of code blocks
    and questions.

    Attributes:
        content_buffer (ContentBuffer): The buffer to which content is streamed.
        tokens (str): Tokens that indicate the end of a statement, for buffer flushing.
        buffer (str): Temporary storage for accumulating streamed tokens.
        stream_finished (threading.Event): Signals when the streaming is finished.
        in_code_block (bool): Indicates whether the current context is a code block.
        in_question_block (bool): Indicates whether the current context is a question block.
        question_buffer (str): Stores the accumulated questions.
    """

    def __init__(self, content_buffer: ContentBuffer, tokens: str = ".?!\n", *args, **kwargs):
        """
        Initializes a new BufferStreamingStdOutCallbackHandler instance.

        Args:
            content_buffer (ContentBuffer): The buffer to which content will be written.
            tokens (str): Tokens that indicate the end of a statement (default: ".?!\n").
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Sets up the callback handler with the given content buffer and tokens.
        Initializes tracking variables for code blocks, buffer content, and the finished signal.
        """
        super().__init__(*args, **kwargs)
        self.content_buffer = content_buffer
        self.tokens = tokens
        self.buffer = ""
        self.stream_finished = threading.Event()

        self.in_code_block = False
        self.in_question_block = False
        self.question_buffer = ""
        log.info("Starting to stream LLM")

    def _is_heartbeat_token(self, token: str) -> bool:
        """Detects if the token is a heartbeat message."""
        return token.startswith('[[HEARTBEAT]]') and token.endswith('[[/HEARTBEAT]]')

    def _strip_heartbeat_markers(self, token: str) -> str:
        """Removes the [[HEARTBEAT]] markers from the token."""
        return token[len('[[HEARTBEAT]]'):-len('[[/HEARTBEAT]]')]

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """
        Processes a new token from the LLM output.

        Args:
            token (str): The new token generated by the LLM.
            **kwargs: Additional keyword arguments.

        Accumulates the token in the buffer and processes it based on the current context.
        The buffer content is written to the content buffer when appropriate tokens or
        patterns are detected.
        """
        #log.debug(f"on_llm_new_token: {token}")

        # Check if the token is a heartbeat message
        if self._is_heartbeat_token(token):
            # Strip the [[HEARTBEAT]] markers and write immediately
            heartbeat_content = self._strip_heartbeat_markers(token)
            #log.info(f"Heartbeat token detected, writing immediately: {heartbeat_content}")
            self.content_buffer.write(heartbeat_content)
        else:
            self.buffer += token

            if '```' in token:
                self.in_code_block = not self.in_code_block

            if not self.in_code_block:
                self._process_buffer()

    def _process_buffer(self):
        """
        Processes the buffer content and writes to the content buffer.

        If the buffer ends with a numbered list pattern or specified tokens, the buffer is flushed
        to the content buffer. Otherwise, the buffer is left intact for further accumulation.
        """
        matches = list(re.finditer(r'\n(\d+\.\s)', self.buffer))
        if matches:
            last_match = matches[-1]
            start_of_last_match = last_match.start() + 1
            self.content_buffer.write(self.buffer[:start_of_last_match])
            self.buffer = self.buffer[start_of_last_match:]
        else:
            if any(self.buffer.endswith(t) for t in self.tokens):
                self.content_buffer.write(self.buffer)
                self.buffer = ""

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """
        Handles the end of LLM streaming.

        Args:
            response (LLMResult): The result returned by the LLM.
            **kwargs: Additional keyword arguments.

        Writes any remaining buffer content to the content buffer, and sets a signal indicating
        that the streaming has finished.
        """
        if self.buffer:
            self.content_buffer.write(self.buffer)
            self.buffer = ""
            log.info("Flushing remaining LLM response buffer")

        self.stream_finished.set()
        log.info("Streaming LLM response ended successfully")



class BufferStreamingStdOutCallbackHandlerAsync(StreamingStdOutCallbackHandler):
    """
    An async callback handler for streaming LLM output to a content buffer.

    This class handles the streaming of output from a large language model (LLM),
    processes tokens from the model output, and writes them to a ContentBuffer.
    It supports handling different types of tokens and keeps track of code blocks
    and questions.

    Attributes:
        content_buffer (ContentBuffer): The buffer to which content is streamed.
        tokens (str): Tokens that indicate the end of a statement, for buffer flushing.
        buffer (str): Temporary storage for accumulating streamed tokens.
        stream_finished (asyncio.Event): Signals when the streaming is finished.
        in_code_block (bool): Indicates whether the current context is a code block.
    """

    def __init__(self, content_buffer: ContentBuffer, tokens: str = ".?!\n", *args, **kwargs):
        """
        Initializes a new BufferStreamingStdOutCallbackHandler instance.

        Args:
            content_buffer (ContentBuffer): The buffer to which content will be written.
            tokens (str): Tokens that indicate the end of a statement (default: ".?!\n").
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Sets up the callback handler with the given content buffer and tokens.
        Initializes tracking variables for code blocks, buffer content, and the finished signal.
        """
        super().__init__(*args, **kwargs)
        self.content_buffer = content_buffer
        self.tokens = tokens
        self.buffer = ""
        self.stream_finished = asyncio.Event()  # Use asyncio.Event for async compatibility
        self.in_code_block = False
        log.info("Starting to stream LLM")

    async def async_on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        #log.debug(f"async_on_llm_new_token: {token}")

        # Check if the token is a heartbeat message
        if self._is_heartbeat_token(token):
            # Strip the [[HEARTBEAT]] markers and write immediately
            heartbeat_content = self._strip_heartbeat_markers(token)
            #log.info(f"Heartbeat token detected, writing immediately: {heartbeat_content}")
            await self.content_buffer.async_write(heartbeat_content)
        else:
            self.buffer += token

            if '```' in token:
                self.in_code_block = not self.in_code_block

            if not self.in_code_block:
                await self._async_process_buffer()

    def _is_heartbeat_token(self, token: str) -> bool:
        """Detects if the token is a heartbeat message."""
        return token.startswith('[[HEARTBEAT]]') and token.endswith('[[/HEARTBEAT]]')

    def _strip_heartbeat_markers(self, token: str) -> str:
        """Removes the [[HEARTBEAT]] markers from the token."""
        return token[len('[[HEARTBEAT]]'):-len('[[/HEARTBEAT]]')]

    async def _async_process_buffer(self):
        """
        Asynchronously processes the buffer content and writes to the content buffer.

        If the buffer ends with a numbered list pattern or specified tokens, the buffer is flushed
        to the content buffer. Otherwise, the buffer is left intact for further accumulation.
        """
        matches = list(re.finditer(r'\n(\d+\.\s)', self.buffer))
        if matches:
            last_match = matches[-1]
            start_of_last_match = last_match.start() + 1
            await self.content_buffer.async_write(self.buffer[:start_of_last_match])
            self.buffer = self.buffer[start_of_last_match:]
        else:
            if any(self.buffer.endswith(t) for t in self.tokens):
                await self.content_buffer.async_write(self.buffer)
                self.buffer = ""

    async def async_on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """
        Asynchronously handles the end of LLM streaming.

        Args:
            response (LLMResult): The result returned by the LLM.
            **kwargs: Additional keyword arguments.
        """
        if self.buffer:
            await self.content_buffer.async_write(self.buffer)
            self.buffer = ""
            log.info("Flushing remaining LLM response buffer")

        self.stream_finished.set()
        log.info("Streaming LLM response ended successfully")




from ..utils import ConfigManager
from ..utils.gcp_project  import get_gcp_project
from ..custom_logging import log
from .discovery_engine_client import DiscoveryEngineClient
from ..components import load_memories
import traceback

def get_all_chunks(question:str, config:ConfigManager):
    """
    Look through a config memory key and find all Vertex AI Search retrievers, call them and return a joined string of chunks

        args: question - question to search similarity for
        config: A ConfigManager object

        returns: a big string of chunks
    """
    memories = load_memories(config=config)
    chunks = []

    if not memories:
        return None

    vector_name = config.vector_name
    for memory in memories:
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore == "discovery_engine" or vectorstore == "vertex_ai_search":
                if value.get('read_only'):
                    new_vector_name = value.get('vector_name')
                    if not new_vector_name:
                        log.warning("read_only specified but no new vector_name to read from")
                        continue
                    else:
                        vector_name = new_vector_name
                
                num_chunks = value.get('num_chunks') or 3

                chunk = get_chunks(question, vector_name, num_chunks)
                if chunk:
                    chunks.append(chunk)
    if chunks:
        return "\n".join(chunks)

    log.warning(f"No chunks found for {vector_name}")
    return None

def get_chunks(question, vector_name, num_chunks):
    de = DiscoveryEngineClient(vector_name, project_id=get_gcp_project(include_config=True))
    try:
        return de.get_chunks(question, num_previous_chunks=num_chunks, num_next_chunks=num_chunks)
    except Exception as err:
        log.error(f"No discovery engine chunks found: {str(err)}")
    


async def async_get_all_chunks(question:str, config:ConfigManager):
    """
    Look through a config memory key and find all Vertex AI Search retrievers, call them and return a joined string of chunks

        args: question - question to search similarity for
        config: A ConfigManager object

        returns: a big string of chunks
    """
    memories = load_memories(config=config)
    chunks = []

    if not memories:
        return None

    vector_name = config.vector_name
    for memory in memories:
        for key, value in memory.items():  # Now iterate over the dictionary
            log.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore')
            if vectorstore == "discovery_engine" or vectorstore == "vertex_ai_search":
                if value.get('read_only'):
                    new_vector_name = value.get('vector_name')
                    if not new_vector_name:
                        log.warning("read_only specified but no new vector_name to read from")
                        continue
                    else:
                        vector_name = new_vector_name
                
                num_chunks = value.get('num_chunks') or 3

                chunk = await async_get_chunks(question, vector_name, num_chunks)
                if chunk:
                    chunks.append(chunk)
    if chunks:
        return "\n".join(chunks)

    log.warning(f"No chunks found for {vector_name}")
    return None

async def async_get_chunks(question, vector_name, num_chunks):
    de = DiscoveryEngineClient(vector_name, project_id=get_gcp_project(include_config=True))
    try:
        return await de.async_get_chunks(question, num_previous_chunks=num_chunks, num_next_chunks=num_chunks)
    except Exception as err:
        log.error(f"No discovery engine chunks found: {str(err)} {traceback.format_exc()}")


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import json
import traceback
import uuid

from ...agents import extract_chat_history, handle_special_commands
from ...qna.parsers import parse_output
from ...streaming import start_streaming_chat
from ...archive import archive_qa
from ...custom_logging import log
from ...utils.config import load_config
from ...utils import ConfigManager
from ...utils.version import sunholo_version
import os
from ...gcs.add_file import add_file_to_gcs, handle_base64_image
from ..swagger import validate_api_key
from datetime import datetime, timedelta

try:
    from flask import request, jsonify, Response
except ImportError:
    pass

try:
    from langfuse.decorators import langfuse_context, observe
except ImportError:
    pass    

# Cache dictionary to store validated API keys
api_key_cache = {}
cache_duration = timedelta(minutes=5)  # Cache duration

def make_openai_response(user_message, vector_name, answer):
    response_id = str(uuid.uuid4())
    log.info(f"openai response: Q: {user_message} to VECTOR_NAME: {vector_name} - A: {answer}")
    openai_response = {
        "id": response_id,
        "object": "chat.completion",
        "created": str(int(datetime.now().timestamp())),
        "model": vector_name,
        "system_fingerprint": sunholo_version(),
        "choices": [{
            "index": 0,
            "message": {
                "role": "assistant",
                "content": answer,
            },
            "logprobs": None,
            "finish_reason": "stop"
        }],
        "usage": {
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "total_tokens": 0
        }
    }

    log.info(f"OpenAI response: {openai_response}")
    return jsonify(openai_response)

def register_qna_routes(app, stream_interpreter, vac_interpreter):
    """
    Register Q&A routes for a Flask application.

    This function sets up multiple routes for handling Q&A operations,
    including streaming responses and processing static responses.

    Args:
        app (Flask): The Flask application instance.
        stream_interpreter (function): Function to handle streaming Q&A responses.
        vac_interpreter (function): Function to handle static Q&A responses.

    Returns:
        None

    Example:
        from flask import Flask
        app = Flask(__name__)

        def dummy_stream_interpreter(...):
            ...

        def dummy_vac_interpreter(...):
            ...

        register_qna_routes(app, dummy_stream_interpreter, dummy_vac_interpreter)
    """
    @app.route("/")
    def home():
        return jsonify("OK")

    @app.route("/health")
    def health():
        return jsonify({"status": "healthy"})

    @app.route('/vac/streaming/<vector_name>', methods=['POST'])
    def stream_qa(vector_name):
        """
        Handle streaming Q&A responses.

        This function sets up a route to handle streaming Q&A responses based on
        the provided vector name.

        Args:
            vector_name (str): The name of the vector for the request.

        Returns:
            Response: A Flask response object streaming the Q&A response content.

        Example:
            response = stream_qa("example_vector")
        """
        observed_stream_interpreter = observe()(stream_interpreter)
        prep = prep_vac(request, vector_name)
        log.debug(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        command_response = prep["command_response"]
        vac_config = prep["vac_config"]
        all_input = prep["all_input"]

        if command_response:
            return jsonify(command_response)

        log.info(f'Streaming data with: {all_input}')
        if span:
            generation = span.generation(
                name="start_streaming_chat",
                metadata=vac_config,
                input = all_input,
                completion_start_time=datetime.now(),
                model=vac_config.get("model") or vac_config.get("llm")
            )

        def generate_response_content():

            for chunk in start_streaming_chat(question=all_input["user_input"],
                                              vector_name=vector_name,
                                              qna_func=observed_stream_interpreter,
                                              chat_history=all_input["chat_history"],
                                              wait_time=all_input["stream_wait_time"],
                                              timeout=all_input["stream_timeout"],
                                              #kwargs
                                              **all_input["kwargs"]
                                              ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    # When we encounter the dictionary, we yield it as a JSON string
                    # and stop the generator.
                    if trace:
                        chunk["trace"] = trace.id
                        chunk["trace_url"] = trace.get_trace_url()
                    archive_qa(chunk, vector_name)
                    if trace:
                        generation.end(output=json.dumps(chunk))
                        span.end(output=json.dumps(chunk))
                        trace.update(output=json.dumps(chunk))
  
                    return json.dumps(chunk)
                
                else:
                    # Otherwise, we yield the plain text chunks as they come in.
                    yield chunk
            
        # Here, the generator function will handle streaming the content to the client.
        response = Response(generate_response_content(), content_type='text/plain; charset=utf-8')
        response.headers['Transfer-Encoding'] = 'chunked'  

        log.debug(f"streaming response: {response}")
        if trace:
            generation.end(output=response)
            span.end(output=response)
            trace.update(output=response)
        
        #if 'user_id' in all_input["kwargs"]:
        #    kwargs = all_input["kwargs"]
        #    config = ConfigManager(vector_name)
        #    add_user_history_rag(kwargs.pop('user_id'), 
        #                         config, 
        #                         question=all_input.pop("user_input"), 
        #                         answer=response.get('answer'), 
        #                         metadata=all_input)

        return response

    @app.route('/vac/<vector_name>', methods=['POST'])
    def process_qna(vector_name):
        """
        Handle static Q&A responses.

        This function sets up a route to handle static Q&A responses based on
        the provided vector name.

        Args:
            vector_name (str): The name of the vector for the request.

        Returns:
            Response: A Flask response object with the Q&A response content.

        Example:
            response = process_qna("example_vector")
        """
        observed_vac_interpreter = observe()(vac_interpreter)
        prep = prep_vac(request, vector_name)
        log.debug(f"Processing prep: {prep}")
        trace = prep["trace"]
        span = prep["span"]
        command_response = prep["command_response"]
        vac_config = prep["vac_config"]
        all_input = prep["all_input"]

        if command_response:
            return jsonify(command_response)

        try:
            if span:
                generation = span.generation(
                    name="vac_interpreter",
                    metadata=vac_config,
                    input = all_input,
                    model=vac_config.get("model") or vac_config.get("llm")
                )
            bot_output = observed_vac_interpreter(
                question=all_input["user_input"],
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                **all_input["kwargs"]
            )
            if span:
                generation.end(output=bot_output)
            # {"answer": "The answer", "source_documents": [{"page_content": "The page content", "metadata": "The metadata"}]}
            bot_output = parse_output(bot_output)
            if trace:
                bot_output["trace"] = trace.id
                bot_output["trace_url"] = trace.get_trace_url()
            archive_qa(bot_output, vector_name)
            log.info(f'==LLM Q:{all_input["user_input"]} - A:{bot_output}')


        except Exception as err: 
            bot_output = {'answer': f'QNA_ERROR: An error occurred while processing /vac/{vector_name}: {str(err)} traceback: {traceback.format_exc()}'}
        
        if trace:
            span.end(output=jsonify(bot_output))
            trace.update(output=jsonify(bot_output)) 

        # {'answer': 'output'}
        return jsonify(bot_output)

    @app.before_request
    def check_authentication_header():
        if request.path.startswith('/openai/'):
            log.debug(f'Request headers: {request.headers}')
            # the header forwarded
            auth_header = request.headers.get('X-Forwarded-Authorization')
            if auth_header:
                
                if auth_header.startswith('Bearer '):
                    api_key = auth_header.split(' ')[1]  # Assuming "Bearer <api_key>"
                else:
                    return jsonify({'error': 'Invalid authorization header does not start with "Bearer " - got: {auth_header}'}), 401
                
                endpoints_host = os.getenv('_ENDPOINTS_HOST')
                if not endpoints_host:
                    return jsonify({'error': '_ENDPOINTS_HOST environment variable not found'}), 401
                
                # Check cache first
                current_time = datetime.now()
                if api_key in api_key_cache:
                    cached_result, cache_time = api_key_cache[api_key]
                    if current_time - cache_time < cache_duration:
                        if not cached_result:
                            return jsonify({'error': 'Invalid cached API key'}), 401
                        else:
                            return  # Valid API key, continue to the endpoint
                    else:
                        # Cache expired, remove from cache
                        del api_key_cache[api_key]
                
                # Validate API key
                is_valid = validate_api_key(api_key, endpoints_host)
                # Update cache
                api_key_cache[api_key] = (is_valid, current_time)
                
                if not is_valid:
                    return jsonify({'error': 'Invalid API key'}), 401
            else:
                return jsonify({'error': 'Missing Authorization header'}), 401

    @app.route('/openai/health', methods=['GET', 'POST'])
    def openai_health_endpoint():
        return jsonify({'message': 'Success'})

    @app.route('/openai/v1/chat/completions', methods=['POST'])
    @app.route('/openai/v1/chat/completions/<vector_name>', methods=['POST'])
    def openai_compatible_endpoint(vector_name=None):
        """
        Handle OpenAI-compatible chat completions.

        This function sets up routes to handle OpenAI-compatible chat completion requests,
        both with and without a specified vector name.

        Args:
            vector_name (str, optional): The name of the vector for the request. Defaults to None.

        Returns:
            Response: A Flask response object with the chat completion content.

        Example:
            response = openai_compatible_endpoint("example_vector")
        """
        data = request.get_json()
        log.info(f'openai_compatible_endpoint got data: {data} for vector: {vector_name}')

        vector_name = vector_name or data.pop('model', None)
        messages = data.pop('messages', None)
        chat_history = data.pop('chat_history', None)
        stream = data.pop('stream', False)

        if not messages:
            return jsonify({"error": "No messages provided"}), 400

        user_message = None
        image_uri = None
        mime_type = None

        for msg in reversed(messages):
            if msg['role'] == 'user':
                if isinstance(msg['content'], list):
                    for content_item in msg['content']:
                        if content_item['type'] == 'text':
                            user_message = content_item['text']
                        elif content_item['type'] == 'image_url':
                            base64_data = content_item['image_url']['url']
                            image_uri, mime_type = handle_base64_image(base64_data, vector_name)
                else:
                    user_message = msg['content']
                break

        if not user_message:
            return jsonify({"error": "No user message provided"}), 400
        else:
            log.info(f"User message: {user_message}")
        
        paired_messages = extract_chat_history(chat_history)
        command_response = handle_special_commands(user_message, vector_name, paired_messages)

        if command_response is not None:

            return make_openai_response(user_message, vector_name, command_response)
        
        if image_uri:
            data["image_uri"] = image_uri
            data["mime"] = mime_type

        all_input = {
            "user_input": user_message,
            "chat_history": chat_history,
            "kwargs": data
        }

        observed_stream_interpreter = observe()(stream_interpreter)

        response_id = str(uuid.uuid4())
        
        def generate_response_content():
            for chunk in start_streaming_chat(question=user_message,
                                            vector_name=vector_name,
                                            qna_func=observed_stream_interpreter,
                                            chat_history=all_input["chat_history"],
                                            wait_time=all_input.get("stream_wait_time", 1),
                                            timeout=all_input.get("stream_timeout", 60),
                                            **all_input["kwargs"]
                                            ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    openai_chunk = {
                        "id": response_id,
                        "object": "chat.completion.chunk",
                        "created": str(int(datetime.now().timestamp())),
                        "model": vector_name,
                        "system_fingerprint": sunholo_version(),
                        "choices": [{
                            "index": 0,
                            "delta": {"content": chunk['answer']},
                            "logprobs": None,
                            "finish_reason": None
                        }]
                    }
                    yield json.dumps(openai_chunk) + "\n"
                else:
                    log.info(f"Unknown chunk: {chunk}")

            final_chunk = {
                "id": response_id,
                "object": "chat.completion.chunk",
                "created": str(int(datetime.now().timestamp())),
                "model": vector_name,
                "system_fingerprint": sunholo_version(),
                "choices": [{
                    "index": 0,
                    "delta": {},
                    "logprobs": None,
                    "finish_reason": "stop"
                }]
            }
            yield json.dumps(final_chunk) + "\n"

        if stream:
            log.info("Streaming openai chunks")
            return Response(generate_response_content(), content_type='text/plain; charset=utf-8')

        try:
            observed_vac_interpreter = observe()(vac_interpreter)
            bot_output = observed_vac_interpreter(
                question=user_message,
                vector_name=vector_name,
                chat_history=all_input["chat_history"],
                **all_input["kwargs"]
            )
            bot_output = parse_output(bot_output)

            log.info(f"Bot output: {bot_output}")
            if bot_output:
                return make_openai_response(user_message, vector_name, bot_output.get('answer', ''))
            else:
                return make_openai_response(user_message, vector_name, 'ERROR: could not find an answer')

        except Exception as err:
            log.error(f"OpenAI response error: {str(err)} traceback: {traceback.format_exc()}")
            
            return make_openai_response(user_message, vector_name, f'ERROR: {str(err)}')


def create_langfuse_trace(request, vector_name):
    """
    Create a Langfuse trace for tracking requests.

    This function initializes a Langfuse trace object based on the request headers
    and vector name.

    Args:
        request (Request): The Flask request object.
        vector_name (str): The name of the vector for the request.

    Returns:
        Langfuse.Trace: The Langfuse trace object.

    Example:
        trace = create_langfuse_trace(request, "example_vector")
    """
    try:
        from langfuse import Langfuse
        langfuse = Langfuse()
    except ImportError as err:
        print(f"No langfuse installed for agents.flask.register_qna_routes, install via `pip install sunholo[http]` - {str(err)}")
        
        return None

    user_id = request.headers.get("X-User-ID")
    session_id = request.headers.get("X-Session-ID")
    message_source = request.headers.get("X-Message-Source")

    package_version = sunholo_version()
    tags = [package_version]
    if message_source:
        tags.append(message_source)

    return langfuse.trace(
        name = f"/vac/{vector_name}",
        user_id = user_id,
        session_id = session_id,
        tags = tags,
        release = f"sunholo-v{package_version}"
    )

def prep_vac(request, vector_name):
    """
    Prepare the input data for a VAC request.

    This function processes the incoming request data, extracts relevant
    information, and prepares the data for VAC processing.

    Args:
        request (Request): The Flask request object.
        vector_name (str): The name of the vector for the request.

    Returns:
        dict: A dictionary containing prepared input data and metadata.

    Example:
        prep_data = prep_vac(request, "example_vector")
    """
    #trace = create_langfuse_trace(request, vector_name)
    trace = None
    span = None

    if request.content_type.startswith('application/json'):
        data = request.get_json()
    elif request.content_type.startswith('multipart/form-data'):
        data = request.form.to_dict()
        if 'file' in request.files:
            file = request.files['file']
            if file.filename != '':
                log.info(f"Found file: {file.filename} to upload to GCS")
                try:
                    image_uri, mime_type = handle_file_upload(file, vector_name)
                    data["image_uri"] = image_uri
                    data["image_url"] = image_uri
                    data["mime"] = mime_type
                except Exception as e:
                    log.error(f"Error uploading file: {str(e)}")
            else:
                log.info("No file selected to upload to GCS")
    else:
        log.warning(f"Error uploading file: Unsupported content type {request.content_type}")

    log.info(f"vac/{vector_name} got data: {data}")

    config, _ = load_config("config/llm_config.yaml")
    vac_configs = config.get("vac")
    if vac_configs:
        vac_config = vac_configs.get(vector_name)
        if not vac_config:
            log.warning("Not a local configured VAC, may be a remote config not synced yet")

    if trace and vac_config:
        trace.update(input=data, metadata=vac_config)

    user_input = data.pop('user_input').strip()
    stream_wait_time = data.pop('stream_wait_time', 7)
    stream_timeout = data.pop('stream_timeout', 120)
    chat_history = data.pop('chat_history', None)
    vector_name = data.pop('vector_name', vector_name)

    log.info("Turning chat_history into paired tuples")
    paired_messages = extract_chat_history(chat_history)

    all_input = {'user_input': user_input, 
                 'vector_name': vector_name, 
                 'chat_history': paired_messages, 
                 'stream_wait_time': stream_wait_time,
                 'stream_timeout': stream_timeout,
                 'kwargs': data}

    if trace:
        span = trace.span(
            name="VAC",
            metadata=vac_config,
            input = all_input
        )
    command_response = handle_special_commands(user_input, vector_name, paired_messages)
    if command_response is not None:
        if trace:
            trace.update(output=jsonify(command_response))
    
    return {
        "trace": trace,
        "span": span,
        "command_response": command_response,
        "all_input": all_input,
        "vac_config": vac_config
    }


def handle_file_upload(file, vector_name):
    """
    Handle file upload and store the file in Google Cloud Storage.

    This function saves the uploaded file locally, uploads it to Google Cloud Storage,
    and then removes the local copy.

    Args:
        file (FileStorage): The uploaded file.
        vector_name (str): The name of the vector for the request.

    Returns:
        tuple: A tuple containing the URI of the uploaded file and its MIME type.

    Raises:
        Exception: If the file upload fails.

    Example:
        uri, mime_type = handle_file_upload(file, "example_vector")
    """
    try:
        file.save(file.filename)
        image_uri = add_file_to_gcs(file.filename, vector_name)
        os.remove(file.filename)  # Clean up the saved file
        return image_uri, file.mimetype
    except Exception as e:
        raise Exception(f'File upload failed: {str(e)}')



from typing import Dict, Any
from ..custom_logging import log

try:
    from langfuse.callback import CallbackHandler
except ImportError:
    CallbackHandler = None

from ..utils.version import sunholo_version

def create_langfuse_callback(**kwargs):

    if not CallbackHandler:
        log.warning("No CallbackHandler found, install langfuse? `pip install langfuse`")
        return None

    # TODO: maybe use langfuse.trace here instead later
    langfuse_handler = CallbackHandler(**kwargs)

    return langfuse_handler

def add_langfuse_tracing(
        config: Dict[str, Any],
        request) -> Dict[str, Any]:
    """
    Config modifier function to add a tracing callback
    By @jmaness https://github.com/langchain-ai/langserve/issues/311

    :param config: config dict
    :param request: HTTP request
    :return: updated config
    """


    log.debug(f"add_langfuse_tracing config: {config} {request}")

    if "callbacks" not in config:
        config["callbacks"] = []

    user_id = request.headers.get("X-User-ID")
    session_id = request.headers.get("X-Session-ID")
    message_source = request.headers.get("X-Message-Source")

    tags = [sunholo_version()]
    if message_source:
        tags.append(message_source)

    log.info(f"Adding langfuse tags to trace: {tags}")
    langfuse_handler = create_langfuse_callback(
        user_id = user_id,
        session_id = session_id,
        tags = tags
    )
    config["callbacks"].extend([langfuse_handler])

    log.debug(f"add_langfuse_tracing modified config {config}")
    return config


#add_routes(app, my_chain,
#           path="/my-chain",
#           per_req_config_modifier=add_langfuse_tracing)